/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef MULTIDACC_HH
#define MULTIDACC_HH

#include "Dacc.hh"
#include "Time.hh"
#include "Interval.hh"
#include "ChannelIndex.hh"
#include "MultiList.hh"
#include <string>
#include <iosfwd>
#include <memory>

using namespace std;

class ChanList;   // Hide the channel list from root.
class TSeries;

/**  MultiDacc is a common interface for several Dacc instances that are to be 
  *  accessed  synchronously. MultiDacc methods are used to access multiple 
  *  data channels from multiple frame file lists. MultiDacc has for the most 
  *  part, the same interfaces as a single Dacc instance. When a fill request 
  *  is made, the data are filled with the same times from each input stream 
  *  in sequence. Before each fill operation is executed, the %MultiDacc 
  *  instance skips any data with times before the latest stream start time. 
  *  Channels must be added to the specific stream that contains the named 
  *  channel. A frame_type method is included to select a steam based on the 
  *  frame type information of the first frame file to be read by each stream.
  *  The %MultiDacc instance also keeps an index of channel names that have 
  *  been requested to resolve the stream number to which it was added.
  *  \brief LIGO Multiple stream data access class.
  *  \author  Junwei Cao
  *  \version 0.2; Last modified December 18, 2008
  */
class MultiDacc
{
public:
    ///  Size type definition
    typedef unsigned int size_type;

    ///  Define the FSeries container.
    typedef Dacc::fseries_type fseries_type;

#ifdef __CINT__
    typedef unsigned int uint32_t;
#endif

public:
    //----------------------------------  Constructors and destructors.
    /**  Default (Null) constructor.
      *  \brief Default constructor.
      */
    MultiDacc(void);

    /**  Construct and connect to source (a list of frame file lists).
      *  \brief Construct and open.
      *  \param Source Input frame path.
      *  \param start  Start of processing time.
      */
    explicit MultiDacc(const std::string& Source, const Time& start=Time(0));

    /**  Destructor.
      *  \brief Destructor.
      */
    ~MultiDacc(void);

    /**  Parse the channel specifier to get the Dacc and name. The specifier 
      *  is of the form \c \<frame-type>/\<channel> or \c \<dacc-id>/\<channel>
      *  \brief Parse channel id.
      *  \param chanid Channel specifier
      *  \param daccid Input stream number (returned)
      *  \param chan   Channel name. (returned)
      */
    void channel_id(const std::string& chanid,
		    int& daccid, std::string& chan) const; 
    
    /**  Close all the currently open files. 
      *  Close all open files.
      */
    void close(void);

    /**  Open a file for each initialized Dacc stream.
      *  \brief Open a frame file for each accessor.
      *  \return Zero on success.
      */
    int open(double timeout=-1.0);
  
    //-----------------------------------  Status manipulation
    /**  Set the DaccIn buffer count.
      *  @memo Set the DaccIn Buffer count.
      *  \param N Buffer count for shared memory input.
      */
    void setBuffer(int N);
  
    /**  If the debug flag is set, explanatory messages will be printed 
      *  when an error is encountered.
      *  @memo Set the debug flag.
      *  \param N Debug printout level.
      */
    void setDebug(int N);

    /**  Set or clear the allowMissingFrame flag. If this flag is set, a 
      *  call to fillData will succede (\e i.e the return code will be 0)
      *  even if one of more streams are missing data.
      *  \brief Set the "allow missing frame" flag.
      *  \param id  Stream allowed to have missing frames.
      *  \param yn  New setting for the "allow missing frame" flag.
      */
    void setAllowMissingFrame(int id, bool yn=true);

    /**  Set or clear the allowPartialStride flag. If this flag is set, a 
      *  call to fillData will succede (\e i.e the return code will be 0)
      *  even if one of more streams do not contain a full stride of data.
      *  If a partial stride is read the getFillStride() method will return
      *  the length of data from the shortest non-optional available data.
      *  \brief Set the "allow partial stride" flag.
      *  \param yn  New setting for the "allow partial stride" flag.
      */
    void setAllowPartialStride(bool yn=true);

    /**  Set the online flag if one or more of the streams is an online 
      *  stream.
      *  \brief Test for an online steam.
      */
    void setIfOnline(void);

    /**  Set or clear the IgnoreMissingChannel flag. If this flag is set, a 
      *  call to fillData will succede (\e i.e the return code will be 0)
      *  even if one of more channels are not found.
      *  \brief Set the "Ignore missing channel" flag.
      *  \param yn  New setting for the "Ignore missing channel" flag.
      */
    void setIgnoreMissingChannel(bool yn);

    /**  Set the noWait flag as specified. If noWait is set, the Dacc 
      *  methods that would otherwise wait for online data (e.g. synch, 
      *  fillData) will instead return with error code -8.
      *  @memo Set the noWait flag.
      *  \param now New setting for the nowait flag.
      */
    void setNoWait(bool now=true);

    /**  The default stride is used if a time interval is not specified
      *  when fillData() is invoked.
      *  @memo Define the default Stride.
      *  \param Dt Default data stride length.
      */
    void setStride(Interval Dt);

    /** multi-stream online timeout.
     */
    void setTimeout(double wtime);
    
    //----------------------------------  Channel list manipulation
    /**  The specified channel is added to the list of channels of
      *  the corresponding Dacc.
      *  @memo Add a channel to the request list.
      *  @param Name  Channel name
      *  @param id    Input stream identifier.
      *  @param decim Channel decimation factor.
      *  @param TSptr Optional message pointer address 
      */
    void addChannel(const std::string& Name, int id=-1, int decim=0, TSeries **TSptr=0);
  
    /**  Add an process frequency series to the list. The purpose of the 
      *  method is similar to that of #addChannel# except that the 
      *  channel being searched for is a processed data frequency series.
      *  @memo Add an FSeries data channel to the request list.
      *  @param Name  Name of channel to collect data from.
      *  \param id    Input stream identifier.
      *  @param FSptr Address of an FSeries pointer.
      */
    void addFSeries(const std::string& Name, int id=-1, fseries_type **FSptr=0);

    /**  Add a Processed data channel to the list. The purpose of the method 
      *  and its arguments are identical to those for addChannel() except 
      *  that any search for the channel is limited to Processed data.
      *  @memo Add a processed data channel to the request list.
      *  @param Name   Name of channel to collect data from.
      *  \param id    Input stream identifier.
      *  @param decim  Decimation factor.
      *  @param TSptr  Address of TSeries pointer.
      */
    void addProcessed(const std::string& Name, int id=-1, int decim=0, TSeries **TSptr=0);

    /**  Add an Adc channel to the list. The purpose of the method and its
      *  arguments are identical to those for addChannel() except that any 
      *  search for the channel is limited to the raw ADC data.
      *  @memo Add a raw data channel to the request list.
      *  @param Name   Name of channel to collect data from.
      *  \param id    Input stream identifier.
      *  @param decim  Decimation factor.
      *  @param TSptr  Address of TSeries pointer.
      */
    void addRaw(const std::string& Name, int id=-1, int decim=0, TSeries **TSptr=0);
  
    /**  Add a Simulated data channel to the list. The purpose of the method 
      *  and its arguments are identical to those for addChannel() except 
      *  that any search for the channel is limited to Simulated data.
      *  @memo Add a simulated data channel to the request list.
      *  @param Name   Name of channel to collect data from.
      *  \param id    Input stream identifier.
      *  @param decim  Decimation factor.
      *  @param TSptr  Address of TSeries pointer.
      */
    void addSimulated(const std::string& Name, int id=-1, int decim=0, TSeries **TSptr=0);

    /**  The named channel is removed from the request list.
      *  @memo Remove a channel from the request list.
      *  \param Name Name of channel to be removed.
      */
    void rmChannel(const std::string& Name);
  
    /**  The data pointer for the frame currently in memory is incremented
      *  past the specified time. If the interval is not specified, 
      *  negative or zero, the rest of the data in the current frame are 
      *  skipped. If a discontinuity in the frame times is found, #flush#
      *  stops skipping at the start of the new data segment.
      *  @return Non-zero if an end of file is found.
      *  @memo Skip data.
      *  @param Stride Time interval to be skipped.
      */
    int flush(Interval Stride=Interval(0.0));

    /**  Get the site and frame type from the first frame path in a stream. 
      *  \brief find a stream with the specified frame type
      *  \param id Stream id
      *  \returns Frame type name for the first frame in the id stream.
      */
    std::string frame_name(int id) const;
  
    /**  Get the stream number for the specified site and frame type. The frame 
      *  type is determined by compating the requested type to the site and
      *  frame type of the first frame in the list assuming a frame path of 
      *  the form
      *  \verbatim 
         <directory>/<site>-<frame-type>-<gps>-<length>.<ext> \endverbatim
      *  An online partition of the form \c /online/\<partition> will match
      *  a \a frame_type equal to the partition name. If none of the current
      *  paths match a requested frame type, the function returns -1.
      *  \brief find a stream with the specified frame type
      *  \param frType A frame type if the form \c \<site>-\<frame-type>
      *  \returns A stream id or -1
      */
    int frame_type(const std::string& frType) const;
  
    /**  Get the current debug printout level.
      *  @memo Get the debug flag.
      *  @return Debug print level.
      */
    int getDebug(void);

    /**  Get the total number of frames read by the first input stream.
      *  \brief Total number of frames
      *  \return Total number of frames read.
      */
    long getTotalFrames(void) const;

    /**  Get the time of the next data to be read from the streams.
      *  \brief  Get the Current time.
      *  \return Start time for next data fill.
      */
    Time getCurrentTime(void) const;

    /**  Returns the length of data collected with the last invocation of
      *  fillData. If partial strides are allowed the fill stride will be 
      *  the length of the shortest non-optional data stride.
      *  \brief Get the duration of the last fill.
      *  \return Duration of last fillData() stride.
      */
    Interval getFillStride(void) const;

    /**  Returns the time at which data collection started with the
      *  last invocation of fillData.
      *  @memo Get the start time of the last fill.
      *  @return Start time of last fillData() stride.
      */
    Time getFillTime(void) const;

    /**  Returns the default time interval over which data are to be collected.
      *  \brief Get the stride interval.
      *  \return Default stride interval.
      */
    Interval getStride(void) const;

    /**  Test whether the specified stream may contain missing fraes.
      *  \brief test whether to allow missing or late frames.
      *  \param id stream number to ignore.
      *  \return true if stream is allowed to have missing or late frames.
      */
    bool ignoreMissingFrame(int id) const;

    /**  Data from channels specified by addChannel() are copied to TSeries
      *  objects. The multiple %Dacc objects are first synchronized to ensure 
      *  that all streams are reading from the same time 
      *  \brief Fill the requested TSeries.
      *  \param Stride %Time interval to be read in.
      *  \param start  Start a stride.
      *  \return 
      *  <table>
      *    <tr><td>0</td><td>Successful completion.</td></tr>
      *    <tr><td>-1</td><td>Frame start not contiguous to previous data.</td>
      *        </tr>
      *    <tr><td>-2</td><td>Sample rate incompatible with previous data.</td>
      *        </tr>
      *    <tr><td>-3</td><td>Requested data not found in current frame.</td>
      *        </tr>
      *    <tr><td>-4</td><td>Error reading frame.</td></tr>
      *    <tr><td>-5</td><td>Frame data are not self-consistent.</td></tr>
      *    <tr><td>-6</td><td>TSeries is not allocated.</td></tr>
      *    <tr><td>-7</td><td>Unsupported data type.</td></tr>
      *    <tr><td>-8</td><td>Signal received while reading.</td></tr>
      *    <tr><td>-9</td><td>Invalid data in structure.</td></tr>
      *  </table>
      */
    int fillData(Interval Stride=Interval(0.0), bool start=true);

    /**  Zero all TSeries in the channel list. Allocate sufficient
      *  memory to hold the given sample time.
      *  @memo   Reset channel TSeries.
      *  @param  Dt     Time interval to be used in allocating memory.
      */
    void zeroChans(Interval Dt);

    /**  Copy data to channel TSeries from a single frame.
      *  @memo   Fill the requested TSeries from a single frame.
      *  @param  Offset Offset of start of data from start of fram.
      *  @param  Dt     Time interval to be read in.
      *  \return 
      *  <table>
      *    <tr><td>0</td><td>Successful completion.</td></tr>
      *    <tr><td>-1</td><td>Frame start not contiguous to previous data.</td>
      *        </tr>
      *    <tr><td>-2</td><td>Sample rate incompatible with previous data.</td>
      *        </tr>
      *    <tr><td>-3</td><td>Requested data not found in current frame.</td>
      *        </tr>
      *    <tr><td>-4</td><td>Error reading frame.</td></tr>
      *    <tr><td>-5</td><td>Frame data are not self-consistent.</td></tr>
      *    <tr><td>-6</td><td>TSeries is not allocated.</td></tr>
      *    <tr><td>-7</td><td>Unsupported data type.</td></tr>
      *  </table>
      */
    int fillChans(Interval Offset, Interval Dt);
  
    /**  Seek the specified time. Frames are read from the listed input file(s)
      *  until one containing the specified time is found.
      *  @memo Skip to the specified time.
      *  \param STime Seek time.
      *  @return Zero if seek was successful.
      */
    int seek(Time STime=Time(0, 0));

    /**  The frame(s) that will provide the data for the next read operation 
      *  are read into memory and the current time is set. If more than one
      *  stream is being read, all streams are synchronized to the latest 
      *  time read by any of the streams.
      *  @memo Synchronize the current frame buffers.
      *  @return Zero if successful or -1 if unable to read in a frame.
      */
    int synch(void);

    /**  Test to see is the names channel has been requested..
      *  @memo Test if channel was requested.
      *  @return True if the specified channel has been requested.
      *  @param Name Channel name to be tested.
      */
    bool isChannelRead(const std::string& Name) const;

    /**  A formatted list of requested channels, their decimation factors, 
      *  and latest time copied is printed to the specified output stream. 
      *  \brief List requested channels.
      *  \param out STL output stream to which channels are listed.
      *  \return Output stream reference.
      */
    std::ostream& list(std::ostream& out) const;

    /**  Add multiple streams to the MultiDacc. Each line in the file 
      *  specified by \a name is the path of a file to be used to define 
      *  the input frames for a stream.
      *  \brief Add multiple input streams
      *  \param name File path for list of input streams.
      */
    void addMulti(const std::string& name);

    /**  Add a single stream to the %MultiDacc. Each line in the file 
      *  specified by \a name is the path of one or more input frames
      *  to be read by the new stream. A stream containing a single path 
      *  may be added by prefixing the frame path with an '='.
      *  \brief Add a single input stream.
      *  \param name Input frame path.
      */
    void addSingle(const std::string& name);

    /**  Get the number of Dacc input streams that will read data in
      *  parallel.
      *  \brief Number of input steams.
      *  \return Number of defined Dacc streams.
      */
    size_t getNDacc(void) const;

    /**  Get a pointer to the ith input stream dDacc object.
      *  \brief Get a pointer to a Dacc stream.
      *  \param i Input stream number.
      *  \return Pointer to the specified input stream.
      */
    Dacc* getDacc(unsigned int i);

    /**  Get a pointer to the ith input stream dDacc object.
      *  \brief Get a pointer to a Dacc stream.
      *  \param i Input stream number.
      *  \return Pointer to the specified input stream.
      */
    const Dacc* getDacc(unsigned int i) const;

    /**  Returns a pointer to the last %TSeries filled for the named channel.
      *  \brief Get a pointer to the data.
      *  \param name Channel name string pointer.
      *  \return Pointer to current data %TSeries.
      */
    TSeries* refData(const std::string& name);

    /**  Returns a pointer to the last %TSeries filled for the named channel.
      *  \brief Get a pointer to the data.
      *  \param name Channel name string pointer.
      *  \return Pointer to current data %TSeries.
      */
    const TSeries* refData(const std::string& name) const;

    /**  Returns a pointer to the last frequency series filled for the named 
      *  channel.
      *  \brief Get a pointer to frequency series data.
      *  \param name Channel name string pointer.
      *  \return Pointer to current data frequency series.
      */
    fseries_type* refFData(const std::string& name);

    /**  Write out class-specific statistical information to the specified 
      *  STL output stream.
      *  \brief Write statistics information.
      *  \param out STL output stream to receive statistics data.
      */
    void writeStats(std::ostream& out) const;

private:

    /** Current status of each of the Dacc instances
     */
    class dacc_status {
    public:
	/** Dacc control flag bits.
	 */ 
	enum flag_bit {
	    kIgnore_missing_frame ///< ignore missing strides
	};
    public:
	/** Construct a dacc status instance, and preset the flag-word.
	 */
	dacc_status(Dacc* dacc, int flags=0);
	/** destroy a dacc status instance and the Dacc instance it references.
	 */
	~dacc_status(void) {}

#ifndef __CINT__
	/** dacc status move constructor to allow the vector template to
	  * add/remove an instance without cloning all the Daccs. (Don't 
	  * tell cint something it won't understand.
	  */
	dacc_status(dacc_status&&) = default;
#endif // !defined(__CINT__)

	/** dacc status is also a pointer.
	 */
	Dacc* operator->(void) const;

	/** set the allow missing frame flag.
	 */
	void  allow_missing_frame(bool yorn);

	/** get the Dacc instance pointer.
	 */
	Dacc* get_ptr(void) const;

	/** test the ignore missing frame flag.
	 */
	bool  ignore_missing_frame(void) const;
	
	/**  Get the stream name. This is either set explicitly with set_name() 
	  *  or it is inferred from the first frame file (or online partition)
	  *  path in the file list.
	  *  \brief Get the stream name
	  *  \return Reference to the stream name string.
	  */
	const std::string& name(void) const {return dacc_name;}
	void  set_flag(flag_bit b, bool tf=true);
	void  set_name(const std::string& nm) {dacc_name = nm;} 
	void  set_state(int state) {dacc_state = state;}
	int   state(void) const {return dacc_state;}
	bool  test_flag(flag_bit b) const {return dacc_flags & (1<<b);}
        /**  Seek data from a specified time limiting the execution time to 
 	  *  an interval from the seek time.
 	  *  \brief Seek data from a given to with online latency limit.
 	  *  \param t Sample time to seek.
 	  *  \param dt Maximum online latency defined as the number of seconds
 	  *            after the seek time.
 	  *  \return Dacc return code: 
 	  *          * 0   Success
 	  *          * -4  Frame read error
 	  *          * -8  Signal received while reading data
 	  *          * -10 timeout occurred
 	  */
	int   tseek(const Time& t, double dt);
    private:
	std::string           dacc_name;
#ifndef __CINT__
	std::unique_ptr<Dacc> dacc_ptr;
#else
	Dacc*                 dacc_ptr;
#endif
	unsigned int          dacc_flags;
	int                   dacc_state;
    };

private:
    /** Define the global status/control flags. The default (preset) state 
     *  of each flag is zero/off, and the flag is named for the one/on state.
     */
    enum multidacc_flags {
	kIs_online,                ///< is online flag bit number
	kAllow_partial_stride      ///< allow partial stride flag bit number
    };
    
    /**  Set a multidacc status/control flag.
     *  \param flag Flag bit number to test.
     *  \param yn   boolean state to which the specified bit will be set.
     */
    void set_mdflag(multidacc_flags flag, bool yn=true);

    /** test a  multidacc status/control flag.
     *  \param flag Flag bit number to test.
     *  \returns current boolean state of the specified flag.
     */
    bool test_mdflag(multidacc_flags flag) const;

private:
    typedef std::list<ChannelIndex> ChannelList;
    typedef ChannelList::iterator chan_iter;
    typedef ChannelList::const_iterator const_chan_iter;

    const_chan_iter findChannel(const std::string& Name) const;
    chan_iter       findChannel(const std::string& Name);

    /**  List of requested channels.
      *  @memo List of requested channels.
      */
    ChannelList mChanList;

    int getDaccIndex(const std::string& Name) const;
    MultiList mList;

    Interval  mStride;
    Interval  mFillStride;

    ///  Frame Input object
    typedef std::vector<dacc_status> DaccVector;
    DaccVector mIn;

    double     mTimeout;
    Time       mSynchTime;
    uint32_t   mGlobalFlags;
};

//======================================  Inline methods
inline size_t
MultiDacc::getNDacc(void) const {
    return mIn.size();
}

inline Dacc*
MultiDacc::getDacc(unsigned int i) {
    return mIn[i].get_ptr();
}

inline const Dacc*
MultiDacc::getDacc(unsigned int i) const {
    return mIn[i].get_ptr();
}

inline void
MultiDacc::set_mdflag(multidacc_flags b, bool tf) {
    if (tf) mGlobalFlags |= (1<<b);
    else    mGlobalFlags &= ~(1<<b);
}

inline bool
MultiDacc::test_mdflag(multidacc_flags b) const {
    return (mGlobalFlags & (1<<b)) != 0;
}

//--------------------------------------  C++11 inlines
#ifndef __CINT__

inline Dacc*
MultiDacc::dacc_status::operator->(void) const {
    return dacc_ptr.get();
}

inline Dacc*
MultiDacc::dacc_status::get_ptr(void) const {
    return dacc_ptr.get();
}

#endif // !defined(__CINT__)

//--------------------------------------  Cint compatible inlines
inline
MultiDacc::dacc_status::dacc_status(Dacc* dacc, int flags)
    : dacc_ptr(dacc), dacc_flags(flags), dacc_state(1)
{}

inline void
MultiDacc::dacc_status::set_flag(flag_bit b, bool tf) {
    if (tf) dacc_flags |= (1<<b);
    else    dacc_flags &= ~(1<<b);
}

inline bool
MultiDacc::dacc_status::ignore_missing_frame(void) const {
    return test_flag(kIgnore_missing_frame);
}

inline void
MultiDacc::dacc_status::allow_missing_frame(bool yn) {
    return set_flag(kIgnore_missing_frame, yn);
}

#endif // MULTIDACC_HH
