/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef MULTIFRAME_WRITER_HH
#define MULTIFRAME_WRITER_HH

#include "FrWriter.hh"
#include "Interval.hh"
#include "Time.hh"
#include "frame_name.hh"
#include <string>

/** The Multi-frame writer class is a driver class for FrWriter to write 
 *  multiframe files. The files may be either a fixed length or an alignment
 *  on the multiple of a specified interval. The fixed length mode is the 
 *  default and may be set either by specifying a length in the constructor, 
 *  or with the set_file_length() method. The aligned file mode is set 
 *  with the set_file_alignment() method. If the length argument is zero or
 *  less than the frame length, a file is written after adding the frame.
 */
class multiframe_writer : public FrWriter {
public:
   multiframe_writer(const std::string& name, int run, Interval length=0.0);
   virtual ~multiframe_writer(void);
   /** Close the current frame file. The expected file duration is compared to 
    *  the actual duration, and if they differ, the frame file is renamed.
    *  \brief Close and rename the file.
    */
   void close(void);
   
   /** Open the frame writer instance. The frame name used is as specified 
    *  by the member frame_name instance and the set_file_name() methods.
    *  The initial start time is as specified by the wtart of the current 
    *  frame and the duration is inferred from the length parameter, the 
    *  current frame end time and the alignment mode.
    *  \brief Open a frame stream.
    */
   void open(void);

   /** Rename the frame file to its nominal path if the original name had
    *  an incorrect length field or a temporary file name was used.
    */
   bool rename(void) const;

   /** Set the file name template. The file name should follow the format 
    *  used by the frame_name::parse() method.
    *  \brief Frame file path template.
    *  \param name File path template.
    */ 
   void set_file_name(const std::string& name);

   /** Set the file name template from the individual components. Directory,
    *  prefix and extension should be specified as for the frame_name methods.
    *  \brief Set the frame file name path template.
    *  \param dir Directory name as specified for frame_name::set_directory()
    *  \param prefix Frame type string as specified for frame_name::set_prefix()
    *  \param ext Frame file extension string (defaults to "gwf").
    */
   void set_file_name(const std::string& dir, const std::string& prefix,
		      const std::string& ext="gwf");

   /** Set the fixed frame length. If the file length is set to a value that is
    *  shorter than a frame length, the frame file will be written with the
    *  appropriate length.
    *  \brief Set fixed frame file length.
    *  \param length Number of seconds of data in a multi-frame file. 
    */
   void set_file_length(Interval length);

   /** Request that frame files be written with a specified end-time alignment. 
    *  Frame files will be written that end on GPS times that are a multiple of 
    *  the specified length. In general frame writing will start and end on 
    *  aligned times after an initial unaligned file is written and until data 
    *  are flushed at the end of a data stream. If the file alignment is set to 
    *  a value that is shorter than a frame length, the frame file will be 
    *  written with a single frame of the appropriate length.
    *  \brief Set frame file end-time alignment.
    *  \param align Frame time alignment factor.
    */
   void set_file_alignment(Interval align);

   /** Request that the files be written to a mangled file path and renamed to
    *  the nominal path after the file is closed.
    *  \brief rename file after it is closed.
    *  \param yn Mangle and rename the file path if true.
    */
   void set_rename(bool yn);

   /** Write the frame built by the FrWriter base class. A frame file is 
    *  created/opened if the file has not been created yet and it is 
    *  closed and renamed if a file is open, but should end before the current
    *  frame.
    *  \brief write a frame.
    */
   void write(void);
private:
   /** Calculate the end time for the current frame file.
    *  \brief Current frame file end time.
    *  \return End time of a currently open frame file.
    */
   Time endTime(void) const;
private:
   frame_name  mFrameName;
   Interval    mLength;
   bool        mAligned;
   bool        mRename;
   Time        mFileStart;
   Time        mFileNow;
};

//======================================  inline accessor methods
inline bool
multiframe_writer::rename(void) const {
   return mRename;
}

#endif // !defined(MULTIFRAME_WRITER_HH)
